import glob
import re
import xml.etree.ElementTree as ET

# Mapping member definitions (id) to all informations
member_definitions: dict = {}

# Compound members
compound_members: dict = {}
compound_metadata: dict = {}

# Mapping doxygen IDs to C++ compound names
doxygen_id_to_name: dict = {}


def resolve_type(node):
    type_node = node.find("type")

    if type_node.find("ref") is not None:
        if type_node.text and "vector" in type_node.text:
          return ["std::vector", type_node.find("ref").attrib["refid"]]
        else:
          return type_node.find("ref").attrib["refid"]
    else:
        return type_node.text


def element_to_string(element):
    s = element.text or ""
    for sub_element in element:
        s += ET.tostring(sub_element).decode("utf-8")
    s += element.tail
    return re.sub('<[^<]+?>', '', s)


def parse_compound(compounddef_node: ET.Element):
    global member_definitions, compound_metadata, compound_members

    name = compounddef_node.find("compoundname").text
    compound_kind = compounddef_node.attrib["kind"]
    if compound_kind == "namespace":
        name = "root"
    id = compounddef_node.attrib["id"]
    doxygen_id_to_name[id] = name

    if name not in compound_members:
        compound_members[name] = []

    compound_metadata[name] = {
        "id": id,
        "kind": compound_kind,
        "name": name,
        "brief": element_to_string(compounddef_node.find("briefdescription/para"))
        if compounddef_node.find("briefdescription/para") is not None
        else None,
    }

    # Searching for member definitions
    for member in compounddef_node.findall("sectiondef/memberdef"):
        kind = member.attrib["kind"]
        id = member.attrib["id"]
        static = member.attrib["static"] == "yes"

        # PyIgnore ?
        ignore = False
        sects = member.findall("detaileddescription/para/xrefsect")
        if sects:
            for sect in sects:
                if "pyignore" in sect.attrib["id"]:
                    ignore = True
        if ignore:
            continue

        if kind in ["function", "variable", "namespace"]:
            member_definitions[id] = {
                "kind": kind,
                "name": member.find("name").text,
                "params": [],
                "static": static,
            }
            member_definitions[id]["type"] = resolve_type(member)

            for param in member.findall("param"):
                param_type = resolve_type(param)
                param_name = None
                if param.find("declname") is not None:
                    param_name = param.find("declname").text
                if param.find("defname") is not None:
                    param_name = param.find("defname").text
                def_val = (
                    param.find("defval").text
                    if param.find("defval") is not None
                    else None
                )
                member_definitions[id]["params"].append(
                    {"type": param_type, "name": param_name, "default": def_val}
                )

            # Brief description
            brief = member.find("briefdescription/para")
            if brief is not None:
                member_definitions[id]["brief"] = element_to_string(brief)

            # Detailed description
            member_definitions[id]["detailed"] = []
            for entry in member.findall(
                "detaileddescription/para/parameterlist/parameteritem"
            ):
                param_name = entry.find("parameternamelist/parametername").text
                param_desc = element_to_string(entry.find("parameterdescription/para"))
                member_definitions[id]["detailed"].append(
                    {"name": param_name, "desc": param_desc}
                )

            # Verbatim
            verbatim = member.find("detaileddescription/para/verbatim")
            if verbatim is not None:
                member_definitions[id]["verbatim"] = verbatim.text

            # Return type
            return_type = member.find(
                "detaileddescription/para/simplesect[@kind='return']"
            )
            if return_type:
                member_definitions[id]["returns"] = element_to_string(return_type.find("para"))

            if compound_kind == "namespace":
                compound_members[name].append(id)

        elif kind == "typedef":
            doxygen_id_to_name[id] = resolve_type(member)
        elif kind == "enum":
            doxygen_id_to_name[id] = name + "::" + member.find("name").text

    # Listing all members
    kind = compounddef_node.attrib["kind"]
    for member in compounddef_node.findall("listofallmembers/member"):
        compound_members[name].append(member.attrib["refid"])


def parse_xml(xml_file: str):
    with open(xml_file, "r") as f:
        xml_content = f.read()
        xml_content = ET.fromstring(xml_content)

        # Searching for compounddef nodes (classs & struct)
        for compounddef_node in xml_content.findall("compounddef"):
            parse_compound(compounddef_node)


def resolve_doxygen_id(id: list|str):
    if type(id) == list:
        tpl = resolve_doxygen_id(id[0])
        typ = resolve_doxygen_id(id[1])
        return f"{tpl}<{typ}>"
    elif id in doxygen_id_to_name:
        return doxygen_id_to_name[id]
    else:
        return id


def parse_directory(directory):
    global member_definitions, compound_members
    doxygen_xml_files = glob.glob(f"{directory}/xml/*.xml")

    for xml_file in doxygen_xml_files:
        parse_xml(xml_file)

    # Resolving types
    for member in member_definitions:
        member_definition = member_definitions[member]
        member_definition["type"] = resolve_doxygen_id(member_definition["type"])

        for param in member_definition["params"]:
            param["type"] = resolve_doxygen_id(param["type"])

    # Resolving compound members
    for name in compound_members:
        temp = compound_members[name]
        compound_members[name] = {}
        for id in temp:
            if id in member_definitions:
                compound_members[name][
                    member_definitions[id]["name"]
                ] = member_definitions[id]


def get_members(name: str):
    if name not in compound_members:
        return None

    return compound_members[name]


def get_metadata(name: str):
    if name not in compound_metadata:
        return None

    return compound_metadata[name]



#!/usr/bin/env python
import re
import inspect
import sys
import os
import argparse

# Current script directory:
repo_directory = os.path.dirname(os.path.realpath(__file__))

# If placo.pyi file already exists next to stubs.py, we read it directly. This is a way to
# avoid running Doxygen when building placo sdist release.
if os.path.exists(f"{repo_directory}/placo.pyi"):
    with open(f"{repo_directory}/placo.pyi", "r") as f:
        print(f.read())
    exit(0)

# Prepending current directory to PYTHONPATH
sys.path = ["."] + sys.path

import placo
from doxygen_parse import parse_directory, get_members, get_metadata

module: str = "placo"

# Ensure Doxygen is run
if not os.path.exists(f"/usr/bin/doxygen"):
    sys.stderr.write("\n-----------------------\n")
    sys.stderr.write("WARNING: Doxygen is not installed\n")
    sys.stderr.write("         you should run: sudo apt install doxygen\n")
    sys.stderr.write("-----------------------\n\n")
    exit(1)

result = os.system(f"cd {repo_directory} && doxygen 1>&2")

# Read the Doxygen XML file
parse_directory(repo_directory)

# Translation for types from C++ to python
rewrite_types: dict = {
    "std::string": "str",
    "double": "float",
    "int": "int",
    "bool": "bool",
    "void": "None",
    "Eigen::MatrixXd": "numpy.ndarray",
    "Eigen::VectorXd": "numpy.ndarray",
    "Eigen::Matrix3d": "numpy.ndarray",
    "Eigen::Vector3d": "numpy.ndarray",
    "Eigen::Matrix2d": "numpy.ndarray",
    "Eigen::Vector2d": "numpy.ndarray",
    "Eigen::Affine3d": "numpy.ndarray",
}

# Building registry and reverse registry for class names
cxx_registry = placo.get_classes_registry()
py_registry = {"root": "root"}
for entry in cxx_registry:
    rewrite_types[entry] = cxx_registry[entry]
    py_registry[cxx_registry[entry]] = entry

def get_member(class_name: str, member_name: str):
    if class_name in py_registry:
        cxx_name = py_registry[class_name]
        members = get_members(cxx_name)
        if members is not None and member_name in members:
            return members[member_name]

    return None


def cxx_type_to_py(typename: str):
    """
    We apply here some heuristics to rewrite C++ types to Python
    """
    typename_raw = typename
    if typename is not None:
        typename = typename.replace("&", "")
        typename = typename.replace("*", "")
        typename = typename.strip()
        if typename.startswith("const "):
            typename = typename[6:]

        if typename in rewrite_types:
            return rewrite_types[typename]
        elif typename.startswith("std::vector"):
            return "list[" + cxx_type_to_py(typename[12:-1]) + "]"
        else:
            return f"any"

    return None


def parse_doc(name: str, doc: str) -> dict:
    """
    Parses the docstring prototypes produced by Boost.Python to infer types
    """
    definition = {"name": name, "args": [], "returns": "any"}

    prototype = doc.strip().split("\n")[0].strip()
    prototype = prototype.replace("[", "").replace("]", "")

    result = re.fullmatch(r"^(.*?)\((.+)\) -> (.+) :$", prototype)
    if result:
        definition["name"] = result.group(1)
        definition["returns"] = result.group(3)

        args = result.group(2).split(",")
        for arg in args:
            arg = arg.strip()
            arg = arg[1:]
            arg_type = ")".join(arg.split(")")[:-1])
            arg_name = arg.split(")")[-1]
            definition["args"].append([arg_type, arg_name])

    return definition


def print_def_prototype(
    method_name: str,
    args: list,
    return_type: str = "any",
    doc="",
    prefix: str = "",
    static: bool = False,
):
    str_definition = ""

    if static:
        str_definition += f"{prefix}@staticmethod\n"

    str_definition += f"{prefix}def {method_name}(\n"
    for arg_name, arg_type, defvalue, comment in args:
        str_definition += f"{prefix}  {arg_name}: {arg_type}"
        if defvalue is not None:
            if arg_type in ["str", "float", "int", "bool"]:
                str_definition += f" = {defvalue.capitalize()}"
        str_definition += ","
        if comment is not None:
            str_definition += f" # {comment}"
        str_definition += "\n"
    str_definition += f"\n{prefix}) -> {return_type}:\n"
    if doc != "":
        str_definition += f'{prefix}  """{doc.strip()}"""\n'
    str_definition += f"{prefix}  ...\n"
    print(str_definition)


def print_def(name: str, doc: str, prefix: str = ""):
    definition = parse_doc(name, doc)

    print_def_prototype(
        definition["name"],
        [[arg_name, arg_type, None, None] for arg_type, arg_name in definition["args"]],
        definition["returns"],
        prefix=prefix,
    )


def print_class_member(class_name: str, member_name: str):
    member = get_member(class_name, member_name)

    if member is not None:
        py_type = cxx_type_to_py(member["type"])
        print(f"  {member_name}: {py_type} # {member['type']}")
        if "brief" in member:
            print(f'  """{member["brief"]}"""')
    else:
        print(f"  {member_name}: any")

    print("")


def print_class_method(class_name: str, method_name: str, doc: str, prefix: str = ""):
    if method_name == "__init__":
        member = get_member(class_name, class_name)
    else:
        member = get_member(class_name, method_name)

    if member is not None:
        static = member["static"]

        # Method arguments
        args = [
            [arg["name"], cxx_type_to_py(arg["type"]), arg["default"], arg["type"]]
            for arg in member["params"]
        ]
        if class_name != "root" and not member["static"]:
            args = [["self", class_name, None, None]] + args

        # Return type
        return_type = "any"
        if member["type"]:
            return_type = cxx_type_to_py(member["type"])

        # Brief
        doc = ""
        if "brief" in member and member["brief"] is not None:
            doc = member["brief"] + "\n"

        if "detailed" in member:
            for param in member["detailed"]:
                doc += f"\n{prefix}  :param {param['name']}: {param['desc']}"

        if "verbatim" in member:
            doc += f"\n{prefix}  {member['verbatim']}"

        if "returns" in member:
            doc += f"\n{prefix}  :return: {member['returns']}"

        print_def_prototype(
            method_name, args, return_type, doc=doc, prefix=prefix, static=static
        )
    else:
        print_def(method_name, doc, prefix)


print("import numpy")
print("import typing")

for name, object in inspect.getmembers(placo):
    if isinstance(object, type):
        class_name = object.__name__
        print(f"{class_name} = typing.NewType(\"{class_name}\", None)")

groups = {}

for name, object in inspect.getmembers(placo):
    if isinstance(object, type):
        class_name = object.__name__
        print(f"class {class_name}:")

        if class_name in py_registry:
            metadata = get_metadata(py_registry[class_name])

            if metadata is not None:
              namespace = "::".join(metadata['name'].split("::")[:-1][:2])
              if namespace not in groups:
                  groups[namespace] = []
              groups[namespace].append(class_name)

            if (
                metadata is not None
                and "brief" in metadata
                and metadata["brief"] is not None
            ):
                print(f"  \"\"\"{metadata['brief']}\"\"\"")

        for _name, _object in inspect.getmembers(object):
            if not _name.startswith("_") or _name == "__init__":
                if callable(_object):
                    print_class_method(class_name, _name, _object.__doc__, "  ")
                else:
                    print_class_member(class_name, _name)
        print("")
    elif callable(object):
        print_class_method("root", name, object.__doc__)
        print("")
    else:
        ...

print(f"__groups__ = {groups}")


import meshcat
import pinocchio as pin
import numpy as np
import meshcat.geometry as g
import meshcat.transformations as tf
import placo


viewer = None
robot_names: dict = {}


def get_viewer() -> meshcat.Visualizer:
    """
    Gets the meshcat viewer, if it doesn't exist, create it
    """
    global viewer

    if viewer is None:
        viewer = meshcat.Visualizer()

        print(f"Viewer URL: {viewer.url()}")

    return viewer


def robot_viz(
    robot: placo.RobotWrapper, name: str = "robot"
) -> pin.visualize.MeshcatVisualizer:
    """
    Builds an instance of pinocchio MeshcatVisualizer, which allows to push the model to the meshcat
    visualizer passed as parameter

    The robot can further be displayed using:

    > viz.display(q)
    """
    global robot_names

    robot_names[robot] = name
    viz = pin.visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    viz.initViewer(viewer=get_viewer())
    viz.loadViewerModel(name)

    return viz


cylinders: dict = {}


def frame_viz(name: str, T: np.ndarray, opacity: float = 1.0) -> None:
    """
    Visualizes a given frame
    """
    global cylinders
    vis = get_viewer()
    axises = {
        "x": (0xFF0000, (-np.pi / 2, [0, 0, 1])),
        "y": (0x00FF00, (0, [0, 1, 0])),
        "z": (0x0000FF, (np.pi / 2, [1, 0, 0])),
    }

    for axis_name in axises:
        color, rotate = axises[axis_name]
        node_name = f"frames/{name}/{axis_name}"

        if node_name not in cylinders:
            cylinders[node_name] = vis[node_name]
            cylinders[node_name].set_object(
                g.Cylinder(0.1, 0.005),
                g.MeshLambertMaterial(color=color, opacity=opacity),
            )
        obj = cylinders[node_name]

        obj.set_transform(
            T @ tf.rotation_matrix(*rotate) @ tf.translation_matrix([0, 0.05, 0])
        )


def point_viz(
    name: str,
    point: np.ndarray,
    radius: float = 0.01,
    color: float = 0xFF0000,
    opacity: float = 1.0,
) -> None:
    """
    Prints a point (sphere)
    """
    vis = get_viewer()
    vis["point"][name].set_object(
        g.Sphere(radius), g.MeshPhongMaterial(color=color, opacity=opacity)
    )
    vis["point"][name].set_transform(tf.translation_matrix(point))


points_sizes = {}


def points_viz(
    name: str,
    points: np.ndarray,
    radius: float = 0.01,
    color: float = 0xFF0000,
    opacity: float = 1.0,
) -> None:
    """
    Prints a point (sphere)
    """
    global points_sizes
    vis = get_viewer()
    k = 0
    for point in points:
        entry = f"{name}_{k}"
        k += 1
        vis["points"][name][entry].set_object(
            g.Sphere(radius), g.MeshPhongMaterial(color=color, opacity=opacity)
        )
        vis["points"][name][entry].set_transform(tf.translation_matrix(point))
    while k < points_sizes.get(name, 0):
        vis["points"][name][f"{name}_{k}"].delete()
        k += 1
    points_sizes[name] = len(points)


def robot_frame_viz(robot: placo.RobotWrapper, frame: str) -> None:
    """
    Draw a frame from the robot
    """
    node_name = f"{robot_names[robot]}_{frame}"
    frame_viz(node_name, robot.get_T_world_frame(frame))


steps: int = 0


def footsteps_viz(footsteps: placo.Footsteps, T: np.ndarray = np.eye(4)) -> None:
    global steps
    vis = get_viewer()

    if len(footsteps) < steps:
        vis["footsteps"].delete()
    steps = len(footsteps)

    k = 0
    for footstep in footsteps:
        k += 1
        polygon = [(T @ [*xy, 0, 1])[:3] for xy in footstep.support_polygon()]
        polygon = np.array([*polygon, polygon[-1]])

        if isinstance(footstep, placo.Support):
            if len(footstep.footsteps) >= 2:
                color = 0x1111AA
            else:
                color = (
                    0xFF3333 if str(footstep.footsteps[0].side) == "left" else 0x33FF33
                )
        else:
            color = 0xFF3333 if str(footstep.side) == "left" else 0x33FF33

        vis["footsteps"][str(k)].set_object(
            g.LineLoop(g.PointsGeometry(polygon.T), g.MeshBasicMaterial(color=color))
        )


def line_viz(name: str, points: np.ndarray, color: float = 0xFF0000) -> None:
    """
    Prints a line
    """
    vis = get_viewer()
    vis["lines"][name].set_object(
        g.LineSegments(g.PointsGeometry(points.T), g.LineBasicMaterial(color=color))
    )


def arrow_viz(
    name: str,
    point_from: np.ndarray,
    point_to: np.ndarray,
    color: float = 0xFF0000,
    radius: float = 0.003,
) -> None:
    """
    Prints an arrow
    """
    head_length = radius * 3
    vis = get_viewer()
    length = np.linalg.norm(point_to - point_from)
    length = max(1e-3, length - head_length)

    cylinder = g.Cylinder(length, radius)
    head = g.Cylinder(head_length, 2 * radius, 0.0, 2 * radius)

    T = tf.translation_matrix(point_from)
    if np.linalg.norm(point_to - point_from) > 1e-6:
        T[:3, :3] = placo.rotation_from_axis(
            "y", (point_to - point_from) / np.linalg.norm(point_to - point_from)
        )

    T_cylinder = T @ tf.translation_matrix(np.array([0, length / 2.0, 0.0]))
    T_head = T @ tf.translation_matrix(np.array([0, length + head_length / 2.0, 0.0]))

    vis["arrows"][name]["cylinder"].set_object(
        cylinder, g.MeshBasicMaterial(color=color)
    )
    vis["arrows"][name]["cylinder"].set_transform(T_cylinder)
    vis["arrows"][name]["head"].set_object(head, g.MeshBasicMaterial(color=color))
    vis["arrows"][name]["head"].set_transform(T_head)


previous_contacts: int = 0


def contacts_viz(solver: placo.DynamicsSolver, ratio=0.1, radius=0.005):
    global previous_contacts
    robot = solver.robot
    frames = robot.frame_names()
    k = 0

    for _ in range(solver.count_contacts()):
        contact = solver.get_contact(k)
        if not contact.active:
            continue

        if isinstance(contact, placo.PointContact):
            frame_name = frames[contact.position_task().frame_index]
            T_world_frame = robot.get_T_world_frame(frame_name)
            arrow_viz(
                f"contact_{k}",
                T_world_frame[:3, 3],
                T_world_frame[:3, 3] + contact.wrench * ratio,
                color=0x00FF00,
                radius=radius,
            )
        elif isinstance(contact, placo.Contact6D):
            frame_name = frames[contact.position_task().frame_index]
            T_world_frame = robot.get_T_world_frame(frame_name)
            wrench = T_world_frame[:3, :3] @ contact.wrench[:3]

            if np.linalg.norm(wrench) < 1e-6:
                origin = T_world_frame[:3, 3]
            else:
                origin = T_world_frame[:3, 3] + T_world_frame[:3, :3] @ contact.zmp()

            arrow_viz(
                f"contact_{k}",
                origin,
                origin + wrench * ratio,
                color=0x00FFAA,
                radius=radius,
            )
        elif isinstance(contact, placo.LineContact):
            frame_name = frames[contact.position_task().frame_index]
            T_world_frame = robot.get_T_world_frame(frame_name)
            wrench = T_world_frame[:3, :3] @ contact.wrench[:3]

            if np.linalg.norm(wrench) < 1e-6:
                origin = T_world_frame[:3, 3]
            else:
                origin = T_world_frame[:3, 3] + T_world_frame[:3, :3] @ contact.zmp()

            arrow_viz(
                f"contact_{k}",
                origin,
                origin + wrench * ratio,
                color=0xFF33AA,
                radius=radius,
            )
        elif isinstance(contact, placo.ExternalWrenchContact):
            frame_name = frames[contact.frame_index]
            T_world_frame = robot.get_T_world_frame(frame_name)
            arrow_viz(
                f"contact_{k}",
                T_world_frame[:3, 3],
                T_world_frame[:3, 3] + contact.w_ext[:3] * ratio,
                color=0xFF2222,
                radius=radius,
            )

        k += 1

    vis = get_viewer()
    k_d = k
    while k_d < previous_contacts:
        vis["arrows"][f"contact_{k_d}"].delete()
        k_d += 1
    previous_contacts = k
